languageSelection = {
  i18n: (function() {

    function asyncFetch(file, fn) {
      return new Promise(function(resolve) {
        var xhr = new XMLHttpRequest();
        xhr.open("GET", chrome.extension.getURL(file), true);
        xhr.onreadystatechange = function() {
          if(this.readyState == 4 && this.responseText != "") {
            resolve(this.responseText);
          }
        };
        try {
          xhr.send();
        }
        catch (e) {
          // File not found, perhaps
        }
      });
    }

    function loadLocales(locales) {
      var messages = {};
      var iterations = [];
      for (var i = 0; i < locales.length; i++) {
        var locale = locales[i];
        var file = "_locales/" + locale + "/messages.json";
        // Doesn't call the callback if file doesn't exist
        var aPromise = (function(aLocale) {
          return asyncFetch(file).then(function(data) {
            messages[aLocale] = JSON.parse(data);
          });
        })(locale);
        iterations.push(aPromise);
      }

      return Promise.all(iterations).then(function(data) {
        return messages;
      });
    }

    function getMessageHelper(messageID, args) {
      if (typeof args == "string")
        args = [args];
      if (l10nData.locales) {
        for (var i = 0; i < l10nData.locales.length; i++) {
          var map = l10nData.messages[l10nData.locales[i]];
          // We must have the locale, and the locale must have the message
          if (map && messageID in map)
            return parseString(map[messageID], args);
        }
      }
      return "";
    }

    // Insert substitution args into a localized string.
    function parseString(msgData, args) {
      // If no substitution, just turn $$ into $ and short-circuit.
      if (msgData.placeholders == undefined && args == undefined)
        return msgData.message.replace(/\$\$/g, '$');

      // Substitute a regex while understanding that $$ should be untouched
      function safesub(txt, re, replacement) {
        var dollaRegex = /\$\$/g, dollaSub = "~~~I18N~~:";
        txt = txt.replace(dollaRegex, dollaSub);
        txt = txt.replace(re, replacement);
        // Put back in "$$" ("$$$$" somehow escapes down to "$$")
        var undollaRegex = /~~~I18N~~:/g, undollaSub = "$$$$";
        txt = txt.replace(undollaRegex, undollaSub);
        return txt;
      }

      var $n_re = /\$([1-9])/g;
      var $n_subber = function(_, num) { return args[num - 1]; };

      var placeholders = {};
      // Fill in $N in placeholders
      for (var name in msgData.placeholders) {
        var content = msgData.placeholders[name].content;
        placeholders[name.toLowerCase()] = safesub(content, $n_re, $n_subber);
      }
      // Fill in $N in message
      var message = safesub(msgData.message, $n_re, $n_subber);
      // Fill in $Place_Holder1$ in message
      message = safesub(message, /\$(\w+?)\$/g, function(full, name) {
        var lowered = name.toLowerCase();
        if (lowered in placeholders)
          return placeholders[lowered];
        return full; // e.g. '$FoO$' instead of 'foo'
      });
      // Replace $$ with $
      message = message.replace(/\$\$/g, '$');

      return message;
    }

    var l10nData = undefined;

    var theI18nObject = {
      // languageSelection.i18n.getMessage() may be used in any extension resource page
      // without any preparation.  But if you want to use it from a content
      // script in Safari, the content script must first run code like this:
      //
      //   get_localization_data_from_global_page_async(function(data) {
      //     languageSelection.i18n._setL10nData(data);
      //     // now I can call languageSelection.i18n.getMessage()
      //   });
      //   // I cannot call getMessage() here because the above call
      //   // is asynchronous.
      //
      // The global page will need to receive your request message, call
      // languageSelection.i18n._getL10nData(), and return its result.
      //
      // We can't avoid this, because the content script can't load
      // l10n data for itself, because it's not allowed to make the xhr
      // call to load the message files from disk.  Sorry :(
      _getL10nData: function(selectedLocale) {
        var result = { locales: [] };
        var availableLocales = getAvailableLocales();

        // == Find all locales we might need to pull messages from, in order
        // A1: If the user has enabled the language selector, the selected 
        //     locale will be passed in as selectedLocale, otherwise
        //     selected locale will be false or undefined
        if (selectedLocale){
          result.locales.push(selectedLocale.replace('-', '_'));
          // A2: Perhaps a region-agnostic version of the selected locale
          if (selectedLocale.length > 2 && availableLocales.indexOf(selectedLocale.substring(0, 2)) != -1)
            result.locales.push(selectedLocale.substring(0, 2));
        } else {
          // B1: The user's current locale, converted to match the format of
          //    the _locales directories (e.g. "en-US" becomes "en_US"
          var currentLocale = navigator.language.replace('-', '_');
          if (availableLocales.indexOf(currentLocale) != -1)
            result.locales.push(currentLocale);
          // B2: Perhaps a region-agnostic version of the current locale
          var simpleCurrentLocale = currentLocale.substring(0,2);
          if (currentLocale.length > 2 && availableLocales.indexOf(simpleCurrentLocale) != -1)
            result.locales.push(simpleCurrentLocale);
        }

        // 3: Set English 'en' as default locale
        if (result.locales.indexOf("en") == -1)
          result.locales.push("en");

        // Load all locale files that exist in that list
        return loadLocales(result.locales).then(function(data) {
          result.messages = data;
          return result;
        });
      },

      // Manually set the localization data.  You only need to call this
      // if using languageSelection.i18n.getMessage() from a content script, before
      // the first call.  You must pass the value of _getL10nData(),
      // which can only be called by the global page.
      _setL10nData: function(data) {
        if (!l10nData) {
          l10nData = { locales:[], messages:{} };
        }
        l10nData.locales = data.locales;
        $.extend(l10nData.messages, data.messages);
      },

      getMessage: function(messageID, args) {
        if (l10nData == undefined) {
          return "";
        }
        return getMessageHelper(messageID, args);
      },

      getActiveLocale: function() {
        if (l10nData && l10nData.messages) {
          return Object.getOwnPropertyNames(l10nData.messages)[0];
        }
        return false;
      }
    };

    return theI18nObject;
  })()
};
